/* 
 * Command to read mac address from otp 
 * For Analog Devices ADSP-SC58x parts with OTP 
 * 
 */



/* 
 * Read OTP on ADSP-SC58x by calling functions in ROM via API.
 * 
 * test code to see if we can all the ROM API
 * 
 */

#include <common.h>
#include <command.h>


//#define COMPILED_IN_LDR_DEBUG
#define READ_LDR_FROM_MEMORY

#ifdef COMPILED_IN_LDR_DEBUG
#include "bs_arm.c"
#endif

#define BLOCK_HEADER_SIZE_BYTES (16)

/* ================================================================================================
        RCU0
   ============================================================================================== */
#define REG_RCU0_CTL                         0x3108B000            /*  RCU0 Control Register */
#define REG_RCU0_STAT                        0x3108B004            /*  RCU0 Status Register */
#define REG_RCU0_CRCTL                       0x3108B008            /*  RCU0 Core Reset Outputs Control Register */
#define REG_RCU0_CRSTAT                      0x3108B00C            /*  RCU0 Core Reset Outputs Status Register */
#define REG_RCU0_SIDIS                       0x3108B010            /*  RCU0 System Interface Disable Register */
#define REG_RCU0_SISTAT                      0x3108B014            /*  RCU0 System Interface Status Register */
#define REG_RCU0_SVECT_LCK                   0x3108B018            /*  RCU0 SVECT Lock Register */
#define REG_RCU0_BCODE                       0x3108B01C            /*  RCU0 Boot Code Register */
#define REG_RCU0_SVECT0                      0x3108B020            /*  RCU0 Software Vector Register 0 */
#define REG_RCU0_SVECT1                      0x3108B024            /*  RCU0 Software Vector Register 1 */
#define REG_RCU0_SVECT2                      0x3108B028            /*  RCU0 Software Vector Register 2 */
#define REG_RCU0_MSG                         0x3108B060            /*  RCU0 Message Register */
#define REG_RCU0_MSG_SET                     0x3108B064            /*  RCU0 Message Set Bits Register */
#define REG_RCU0_MSG_CLR                     0x3108B068            /*  RCU0 Message Clear Bits Register */


/* block code def */
#define BCODE_POS       (0)
#define BCODE_SZ        (4)
#define BFLAG_SAVE_POS  (4)
#define BFLAG_SAVE_SZ   (1)
#define BFLAG_AUX_POS   (5)
#define BFLAG_AUX_SZ    (1)
#define BFLAG_FILL_POS  (8)
#define BFLAG_FILL_SZ   (1)
#define BFLAG_CALLBACK_POS (10)
#define BFLAG_CALLBACK_SZ  (1)
#define BFLAG_INIT_POS     (11)
#define BFLAG_INIT_SZ      (1)
#define BFLAG_IGNORE_POS   (12)
#define BFLAG_IGNORE_SZ    (1)
#define BFLAG_INDIRECT_POS (13)
#define BFLAG_INDIRECT_SZ  (1)
#define BFLAG_FIRST_POS    (14)
#define BFLAG_FIRST_SZ     (1)
#define BFLAG_FINAL_POS    (15)
#define BFLAG_FINAL_SZ     (1)
#define HDRCHK_POS         (16)
#define HDRCHK_SZ          (8)
#define HDRSIGN_POS        (24)
#define HDRSIGN_SZ         (8)

#define CORE_CODE_ARM  (0xAD)
#define CORE_CODE_SH0  (0xAC)
#define CORE_CODE_SH1  (0xAB)


typedef enum{
        block_code_idx=0,
        target_address_idx,
        byte_count_idx,
        argument_idx
} blockHeaderIndex;


void print_block_header(const unsigned long bh[])
{
        printf("**block header**\n");
        printf("block code:     0x%08x\n", bh[block_code_idx]);
        printf("target address: 0x%08x\n", bh[target_address_idx]);
        printf("byte count:     0x%08x\n", bh[byte_count_idx]);
        printf("argument:       0x%08x\n", bh[argument_idx]);
        return;
}


void print_block_payload(const unsigned char buffer[], unsigned long count)
{
        int i;
        printf("\nBlock payload ...\n");
        for(i=0; i<count; i++){
                printf("0x%02x, ", buffer[i]);
        }
        printf("\n");
        return;

}

void process_block_payload(const unsigned char buffer[], unsigned long target_address, unsigned long count, bool fb)
{
	int i;
	unsigned char *tgtmem = (unsigned char *)target_address;

	/* move data to target memory */
	if(fb == false){
		for(i=0; i<count; i++){
			tgtmem[i] = buffer[i];
		}
	}
	else{
		for(i=0; i<count; i++){
			tgtmem[i] = 0;
		}
	}

	return;		
}



int do_bootldr(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{

#if defined(COMPILED_IN_LDR_DEBUG) && !defined(READ_LDR_FROM_MEMORY)
	extern uint8_t bootstream[];
#elif defined(READ_LDR_FROM_MEMORY) && !defined(COMPILED_IN_LDR_DEBUG)
	uint8_t *bootstream;
#endif

        unsigned long blockHdr[4];
        unsigned long blockCnt = 0;
	unsigned long appStartAddr0 = 0;
	unsigned long appStartAddr1 = 0;
	unsigned long appStartAddr2 = 0;
        bool lastBlock = false;
        bool skipProcessPayload = false;
        bool ignoreBuffer = false;
        bool fillBlock = false;
	bool firstBlock = false;
	int i;
        size_t result;
	void (*bootedProg)(void);
	uint8_t *workingBuf; 

#if defined(READ_LDR_FROM_MEMORY) && !defined(COMPILED_IN_LDR_DEBUG)
        if(argc < 2){
                printf("usage: bootldr <address where LDR is stored>\n");
                return -1;
        }
        else{
                printf("opening 0x%s\n", argv[1]);
        }

	bootstream = (uint8_t *)simple_strtol(argv[1], NULL, 16);
#endif
	workingBuf = bootstream;



        do{
                /* print block count */
                printf("Block #%d\n", blockCnt);

		blockHdr[0] = (workingBuf[3]<<24) | (workingBuf[2]<<16) | (workingBuf[1]<<8) | (workingBuf[0]);
		blockHdr[1] = (workingBuf[7]<<24) | (workingBuf[6]<<16) | (workingBuf[5]<<8) | (workingBuf[4]);
		blockHdr[2] = (workingBuf[11]<<24) | (workingBuf[10]<<16) | (workingBuf[9]<<8) | (workingBuf[8]);
		blockHdr[3] = (workingBuf[15]<<24) | (workingBuf[14]<<16) | (workingBuf[13]<<8) | (workingBuf[12]);
		workingBuf += BLOCK_HEADER_SIZE_BYTES;

                /* print out header */
                print_block_header(blockHdr);

                /* last block? */
                if(blockHdr[block_code_idx]  & (1 << BFLAG_FINAL_POS)){
                        lastBlock = true;
                        printf("LAST BLOCK\n");
                }


                /* ignore buffer */
                if(blockHdr[block_code_idx] & (1 << BFLAG_IGNORE_POS)){
                        ignoreBuffer = true;
                        printf("IGNORE BUFFER\n");
                }

                /* fill block */
                if(blockHdr[block_code_idx] & (1 << BFLAG_FILL_POS)){
                        fillBlock = true;
                        printf("FILL BLOCK\n");
                }

		/* first block */
                if(blockHdr[block_code_idx] & (1 << BFLAG_FIRST_POS)){
			uint8_t core_code;
                        firstBlock = true;
                        printf("FIRST BLOCK\n");
			core_code = ( (blockHdr[block_code_idx] >> HDRSIGN_POS) & ((1<<HDRSIGN_SZ)-1)) ;
			if(core_code == CORE_CODE_ARM){
				appStartAddr0 = blockHdr[target_address_idx];
			}
			else if(core_code == CORE_CODE_SH0){
				appStartAddr1 = blockHdr[target_address_idx];
			}
			else if(core_code == CORE_CODE_SH1){
				appStartAddr2 = blockHdr[target_address_idx];
			}
		}	


                if(ignoreBuffer == false && fillBlock == false){
                        /* process payload */
                        if(skipProcessPayload == false){
                                //print_block_payload(workingBuf, blockHdr[byte_count_idx]);
                                process_block_payload(workingBuf, blockHdr[target_address_idx], blockHdr[byte_count_idx], false);
                        }

			workingBuf += blockHdr[byte_count_idx];

                }


		if(fillBlock == true){
                	process_block_payload(workingBuf, blockHdr[target_address_idx], blockHdr[byte_count_idx], true);
		}


                if(lastBlock == true){
                        break;
                }

                blockCnt++;

                /* reset flags */
                ignoreBuffer = false;
                fillBlock = false;
		firstBlock = false;

        }while(1);

        /* set svect and reset */

	printf("\ntarget address (ARM): 0x%x\n", appStartAddr0);
	*((volatile unsigned long *)(REG_RCU0_SVECT0)) = appStartAddr0;

	printf("\ntarget address (SH0): 0x%x\n", appStartAddr1);
	*((volatile unsigned long *)(REG_RCU0_SVECT1)) = appStartAddr1;

	printf("\ntarget address (SH1): 0x%x\n", appStartAddr2);
	*((volatile unsigned long *)(REG_RCU0_SVECT2)) = appStartAddr2;


#if 0
	*((volatile unsigned long *)(REG_RCU0_CRSTAT)) &= 0xfffffffe;
	*((volatile unsigned long *)(REG_RCU0_CRCTL)) |= 1; 

	while( (*((volatile unsigned long *)(REG_RCU0_CRSTAT)) & 1) == 0)
        	;


	*((volatile unsigned long *)(REG_RCU0_CRCTL)) &= 0xfffffffe;
#else
	bootedProg = appStartAddr0;
	bootedProg();
#endif
	

	
	return 0;
}

/* decl protocol below: <name of cmd>, <max # of args including command>, <auto-repeat true/false>, <function for command> */ 
U_BOOT_CMD(
	bootldr, 4, 0,  do_bootldr,
	"boot LDR file \n",
	"boot LDR file \n"
);
